
from .grids import make_axis
from .operators import Diff as _Diff


class Diff(_Diff):

    def __init__(self, axis=0, grid=None, periodic=False, acc=_Diff.DEFAULT_ACC):
        """Represents a partial derivative (along one axis).

        For higher derivatives, exponentiate. For mixed partial derivatives, multiply. See
        examples below.

        Examples
        --------
        Set up grid (equidistant here):
            >>> import numpy as np
            >>> x = np.linspace(0, 10, 100)

        The array to differentiate
            >>> f = np.sin(x) # as an example

        Define the first derivative:
            >>> from findiff import Diff
            >>> d_dx = Diff(0)
            >>> d_dx.set_grid({0: x[1] - x[0]})

        Now apply it:
            >>> df_dx = d_dx(f)

        The second derivative is constructed by exponentiation:
            >>> d2_dx2 = d_dx ** 2
            >>> d2f_dx2 = d2_dx2(f)

        In multiple dimensions with meshed grids, the usage is accordingly:
            >>> x = y = z = np.linspace(0, 10, 100)
            >>> dx = dy = dz = x[1] - x[0]
            >>> X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
            >>> f = np.sin(X) * np.sin(Y) * np.sin(Z)
            >>> d_dx = Diff(0)
            >>> d_dy = Diff(1)
            >>> d_dz = Diff(2)
            >>> d3_dxdydz = d_dx * d_dy * d_dz
            >>> d3_dxdydz.set_grid({0: dx, 1: dy, 2: dz})
            >>> d3f_dxdydz = d3_dxdydz(f)
        """
        grid_axis = make_axis(axis, grid, periodic)
        super().__init__(axis, grid_axis, acc)



import numpy as np
import pytest


@pytest.fixture()
def grid_data(request):
    m = request.node.get_closest_marker("grid_spec")
    if m is None:
        raise ValueError("Grid data marker not defined")
    shape = m.kwargs["shape"]
    edges = m.kwargs["edges"]
    endpoints = m.kwargs.get("endpoints", [True] * len(shape))

    if len(shape) == 1:
        x = np.linspace(*edges, shape[0], endpoint=endpoints[0])
        dx = x[1] - x[0]
        return x, dx

    axes = tuple(
        [
            np.linspace(edges[k][0], edges[k][1], shape[k], endpoint=endpoints[k])
            for k in range(len(shape))
        ]
    )
    coords = np.meshgrid(*axes, indexing="ij")
    spacings = [axes[k][1] - axes[k][0] for k in range(len(shape))]
    return axes, spacings, coords



import numbers
from abc import ABC, abstractmethod

import numpy as np
from scipy import sparse

from findiff.findiff import build_differentiator
from findiff.grids import GridAxis, make_grid
from findiff.stencils import StencilSet


class Expression(ABC):
    """Represents a differential operator expression."""

    __array_priority__ = 100  # Makes sure custom multiplication is called over numpy's

    def __init__(self, *args, **kwargs):
        self.children = []

    @abstractmethod
    def __call__(self, f, *args, **kwargs):
        """Applies the differential operator expression L to array f.

        Parameters
        ----------
        f : numpy.ndarray
            The arrow to differentiate.

        Returns
        -------
        numpy.ndarray
            Array of the same shape as f containing the derivative values.
        """
        pass

    @abstractmethod
    def matrix(self, shape):
        """Returns a matrix representation of the differential operator for a given grid shape."""
        pass

    def stencil(self, shape):
        """Returns a stencil representation of the differential operator for a given grid shape."""
        return StencilSet(self, shape)

    def __add__(self, other):
        """Allows to add differential operator expressions."""
        return Add(self, other)

    def __radd__(self, other):
        """Allows to add differential operator expressions."""
        return Add(self, other)

    def __sub__(self, other):
        """Allows to subtract differential operator expressions."""
        return Add(ScalarOperator(-1) * other, self)

    def __rsub__(self, other):
        """Allows to subtract differential operator expressions."""
        return Add(ScalarOperator(-1) * other, self)

    def __mul__(self, other):
        """Allows to multiply differential operator expressions."""
        return Mul(self, other)

    def __rmul__(self, other):
        """Allows to multiply differential operator expressions."""
        return Mul(other, self)

    @property
    def grid(self):
        """Returns the grid used."""
        return getattr(self, "_grid", None)

    def set_grid(self, grid):
        """Sets the grid for the given differential operator expression.

        Parameters
        ----------
        grid: dict | Grid
            Specifies the grid to use. If a dict is given, an equidistant grid
            is assumed and the dict specifies the spacings along the required axes.
        """
        self._grid = make_grid(grid)
        for child in self.children:
            child.set_grid(self._grid)

    def set_accuracy(self, acc):
        """Sets the requested accuracy for the given differential operator expression.

        Parameters
        ----------
        acc: int
            The accuracy order. Must be a positive, even number.
        """
        self.acc = acc
        for child in self.children:
            child.set_accuracy(acc)


class FieldOperator(Expression):
    """An operator that multiplies an array pointwise."""

    def __init__(self, value):
        super().__init__()
        self.value = value

    def __call__(self, f, *args, **kwargs):
        if isinstance(f, (numbers.Number, np.ndarray)):
            return self.value * f
        return self.value * super().__call__(f, *args, **kwargs)

    def matrix(self, shape):
        if isinstance(self.value, np.ndarray):
            diag_values = self.value.reshape(-1)
            return sparse.diags(diag_values)
        elif isinstance(self.value, numbers.Number):
            siz = np.prod(shape)
            return sparse.diags(self.value * np.ones(siz))


class ScalarOperator(FieldOperator):
    """A multiple of the identity operator."""

    def __init__(self, value):
        if not isinstance(value, numbers.Number):
            raise ValueError("Expected number, got " + str(type(value)))
        super().__init__(value)

    def matrix(self, shape):
        siz = np.prod(shape)
        mat = sparse.lil_matrix((siz, siz))
        diag = list(range(siz))
        mat[diag, diag] = self.value
        return sparse.csr_matrix(mat)


class Identity(ScalarOperator):
    """The identity operator."""

    def __init__(self):
        super().__init__(1)


class BinaryOperation(Expression):

    @property
    def left(self):
        return self.children[0]

    @property
    def right(self):
        return self.children[1]


class Add(BinaryOperation):
    def __init__(self, left, right):
        if isinstance(left, (numbers.Number, np.ndarray)):
            left = FieldOperator(left)
        if isinstance(right, (numbers.Number, np.ndarray)):
            right = FieldOperator(right)
        super().__init__()
        self.children = [left, right]

    def __call__(self, f, *args, **kwargs):
        return self.left(f, *args, **kwargs) + self.right(f, *args, **kwargs)

    def matrix(self, shape):
        return self.left.matrix(shape) + self.right.matrix(shape)


class Mul(BinaryOperation):
    def __init__(self, left, right):
        if isinstance(left, (numbers.Number, np.ndarray)):
            left = FieldOperator(left)
        if isinstance(right, (numbers.Number, np.ndarray)):
            right = FieldOperator(right)
        super().__init__()
        self.children = [left, right]

    def __call__(self, f, *args, **kwargs):
        return self.left(self.right(f, *args, **kwargs), *args, **kwargs)

    def matrix(self, shape):
        return self.left.matrix(shape) * self.right.matrix(shape)


class Diff(Expression):

    DEFAULT_ACC = 2

    def __init__(self, dim, axis: GridAxis = None, acc=DEFAULT_ACC):
        """Initializes a Diff instance.

        Parameters
        ----------
        dim: int
            The 0-based index of the axis along which to take the derivative.
        axis: GridAxis
            The grid axis.
        acc: (optional) int
            The accuracy order to use. Must be a positive even number.
        """
        super().__init__()

        self.set_axis(axis)
        self.dim = dim
        self.acc = acc
        self._order = 1
        self._differentiator = None

    def set_grid(self, grid):
        super().set_grid(grid)
        self.set_axis(self.grid.get_axis(self.dim))

    def set_axis(self, axis: GridAxis):
        self._axis = axis
        self._differentiator = None

    @property
    def axis(self):
        return self._axis

    @property
    def order(self):
        """Returns the order of the derivative."""
        return self._order

    def __call__(self, f, *args, **kwargs):
        """Applies the differential operator."""

        if "acc" in kwargs:
            # allow to pass down new accuracy deep in expression tree
            new_acc = kwargs["acc"]
            if new_acc != self.acc:
                self._differentiator = None
                self.set_accuracy(new_acc)

        if isinstance(f, Expression):
            f = f(*args, **kwargs)

        return self.differentiator(f)

    @property
    def differentiator(self):
        if self._differentiator is None:
            self._differentiator = build_differentiator(self.order, self.axis, self.acc)
        return self._differentiator

    def matrix(self, shape):
        return self.differentiator.matrix(shape)

    def __pow__(self, power):
        """Returns a Diff instance for a higher order derivative."""
        new_diff = Diff(self.dim, self.axis, acc=self.acc)
        new_diff._order *= power
        return new_diff

    def __mul__(self, other):
        if isinstance(other, Diff) and self.dim == other.dim:
            new_diff = Diff(self.dim, self.axis, acc=self.acc)
            new_diff._order += other.order
            return new_diff
        return super().__mul__(other)



"""A module for the common differential operators of vector calculus"""

import numpy as np

from .compatible import FinDiff


class VectorOperator:
    """Base class for all vector differential operators.
    Shall not be instantiated directly, but through the child classes.
    """

    def __init__(self, **kwargs):
        """Constructor for the VectorOperator base class.

        kwargs:
        -------

        h       list with the grid spacings of an N-dimensional uniform grid

        coords  list of 1D arrays with the coordinate values along the N axes.
                This is used for non-uniform grids.

        Either specify "h" or "coords", not both.

        """

        if "acc" in kwargs:
            self.acc = kwargs.pop("acc")
        else:
            self.acc = 2

        if (
            "spac" in kwargs or "h" in kwargs
        ):  # necessary for backward compatibility 0.5.2 => 0.6
            if "spac" in kwargs:
                kw = "spac"
            else:
                kw = "h"
            self.h = kwargs.pop(kw)
            self.ndims = len(self.h)
            self.components = [FinDiff(k, self.h[k], 1) for k in range(self.ndims)]

        if "coords" in kwargs:
            coords = kwargs.pop("coords")
            self.ndims = self.__get_dimension(coords)
            self.components = [
                FinDiff((k, coords[k], 1), **kwargs) for k in range(self.ndims)
            ]

    def __get_dimension(self, coords):
        return len(coords)


class Gradient(VectorOperator):
    r"""
    The N-dimensional gradient.

    .. math::
        \nabla = \left(\frac{\partial}{\partial x_0}, \frac{\partial}{\partial x_1}, ... , \frac{\partial}{\partial x_{N-1}}\right)

    :param kwargs:  exactly one of *h* and *coords* must be specified

             *h*
                     list with the grid spacings of an N-dimensional uniform grid
             *coords*
                     list of 1D arrays with the coordinate values along the N axes.
                     This is used for non-uniform grids.

             *acc*
                     accuracy order, must be positive integer, default is 2
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, f):
        """
        Applies the N-dimensional gradient to the array f.

        :param f:  ``numpy.ndarray``

                Array to apply the gradient to. It represents a scalar function,
                so it must have N axes for the N independent variables.

        :returns: ``numpy.ndarray``

                The gradient of f, which has N+1 axes, i.e. it is
                an array of N arrays of N axes each.

        """

        if not isinstance(f, np.ndarray):
            raise TypeError("Function to differentiate must be numpy.ndarray")

        if len(f.shape) != self.ndims:
            raise ValueError("Gradients can only be applied to scalar functions")

        result = []
        for k in range(self.ndims):
            d_dxk = self.components[k]
            result.append(d_dxk(f, acc=self.acc))

        return np.array(result)


class Divergence(VectorOperator):
    r"""
    The N-dimensional divergence.

    .. math::

       {\rm \bf div} = \nabla \cdot

    :param kwargs:  exactly one of *h* and *coords* must be specified

         *h*
                 list with the grid spacings of an N-dimensional uniform grid
         *coords*
                 list of 1D arrays with the coordinate values along the N axes.
                 This is used for non-uniform grids.

         *acc*
                 accuracy order, must be positive integer, default is 2

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, f):
        """
        Applies the divergence to the array f.

        :param f: ``numpy.ndarray``

               a vector function of N variables, so its array has N+1 axes.

        :returns: ``numpy.ndarray``

               the divergence, which is a scalar function of N variables, so it's array dimension has N axes

        """
        if not isinstance(f, np.ndarray) and not isinstance(f, list):
            raise TypeError(
                "Function to differentiate must be numpy.ndarray or list of numpy.ndarrays"
            )

        if len(f.shape) != self.ndims + 1 and f.shape[0] != self.ndims:
            raise ValueError(
                "Divergence can only be applied to vector functions of the same dimension"
            )

        result = np.zeros(f.shape[1:])

        for k in range(self.ndims):
            result += self.components[k](f[k], acc=self.acc)

        return result


class Curl(VectorOperator):
    r"""
    The curl operator.

    .. math::

        {\rm \bf rot} = \nabla \times

    Is only defined for 3D.

    :param kwargs:  exactly one of *h* and *coords* must be specified

     *h*
             list with the grid spacings of a 3-dimensional uniform grid
     *coords*
             list of 1D arrays with the coordinate values along the 3 axes.
             This is used for non-uniform grids.

     *acc*
             accuracy order, must be positive integer, default is 2


    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.ndims != 3:
            raise ValueError(
                f"Curl operation is only defined in 3 dimensions. {self.ndims} were given."
            )

    def __call__(self, f):
        """
        Applies the curl to the array f.

        :param f: ``numpy.ndarray``

               a vector function of N variables, so its array has N+1 axes.

        :returns: ``numpy.ndarray``

               the curl, which is a vector function of N variables, so it's array dimension has N+1 axes

        """

        if not isinstance(f, np.ndarray) and not isinstance(f, list):
            raise TypeError(
                "Function to differentiate must be numpy.ndarray or list of numpy.ndarrays"
            )

        if len(f.shape) != self.ndims + 1 and f.shape[0] != self.ndims:
            raise ValueError(
                "Curl can only be applied to vector functions of the three dimensions"
            )

        result = np.zeros(f.shape)

        result[0] += self.components[1](f[2], acc=self.acc) - self.components[2](
            f[1], acc=self.acc
        )
        result[1] += self.components[2](f[0], acc=self.acc) - self.components[0](
            f[2], acc=self.acc
        )
        result[2] += self.components[0](f[1], acc=self.acc) - self.components[1](
            f[0], acc=self.acc
        )

        return result


class Laplacian:
    r"""
    The N-dimensional Laplace operator.

    .. math::

       {\rm \bf \nabla^2} = \sum_{k=0}^{N-1} \frac{\partial^2}{\partial x_k^2}

    :param kwargs:  exactly one of *h* and *coords* must be specified

         *h*
                 list with the grid spacings of an N-dimensional uniform grid
         *coords*
                 list of 1D arrays with the coordinate values along the N axes.
                 This is used for non-uniform grids.

         *acc*
                 accuracy order, must be positive integer, default is 2

    """

    """A representation of the Laplace operator in arbitrary dimensions using finite difference schemes"""

    def __init__(self, h=None, acc=2):
        h = h or [1.0]
        h = wrap_in_ndarray(h)

        self._parts = [FinDiff((k, h[k], 2), acc=acc) for k in range(len(h))]

    def __call__(self, f):
        """
        Applies the Laplacian to the array f.

        :param f: ``numpy.ndarray``

               a scalar function of N variables, so its array has N axes.

        :returns: ``numpy.ndarray``

               the Laplacian of f, which is a scalar function of N variables, so it's array dimension has N axes

        """
        laplace_f = np.zeros_like(f)

        for part in self._parts:
            laplace_f += part(f)

        return laplace_f


def wrap_in_ndarray(value):
    """Wraps the argument in a numpy.ndarray.

    If value is a scalar, it is converted in a list first.
    If value is array-like, the shape is conserved.

    """

    if hasattr(value, "__len__"):
        return np.array(value)
    else:
        return np.array([value])



import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from findiff import Diff

pytestmark = pytest.mark.functional


@pytest.mark.one_dimensional
class TestDiff_1D:

    @pytest.mark.grid_spec(shape=(101,), edges=(0, 1))
    def test_diff(self, grid_data):
        x, dx = grid_data

        u = x**2
        expected = 2 * x

        fd = Diff(0, dx)
        actual = fd(u)

        assert_array_almost_equal(expected, actual)

    @pytest.mark.grid_spec(shape=(101,), edges=(0, 1))
    def test_diff_deferred(self, grid_data):
        x, dx = grid_data

        u = x**2
        expected = 2 * x

        fd = Diff(0)
        fd.set_grid({0: dx})
        actual = fd(u)

        assert_array_almost_equal(expected, actual)

    def test_diff_deferred_called_too_early(self):
        u = np.zeros(10)
        fd = Diff(0)
        with pytest.raises(TypeError, match="Unknown axis type."):
            fd(u)

    @pytest.mark.default_args
    @pytest.mark.grid_spec(shape=(101,), edges=(0, 1))
    def test_diff_defaults(self, grid_data):
        x, dx = grid_data

        u = x**2
        expected = 2 * x

        fd = Diff()
        fd.set_grid({0: dx})
        actual = fd(u)

        assert_array_almost_equal(expected, actual)

    @pytest.mark.invalid_args
    def test_diff_invalid_args(self):

        with pytest.raises(ValueError, match="Dimension must be >= 0"):
            Diff(-1, 1)

        with pytest.raises(ValueError, match="Spacing must be > 0."):
            Diff(0, -1)

    @pytest.mark.periodic
    @pytest.mark.grid_spec(shape=[200], edges=(0, 2 * np.pi), endpoints=[False])
    def test_diff_periodic(self, grid_data):
        x, dx = grid_data
        f = np.sin(x)

        actual = Diff(0, dx, periodic=True, acc=4)(f)

        assert_array_almost_equal(np.cos(x), actual)



import numpy as np

from findiff import Diff


def test_iterative_accuracy():

    ns = np.logspace(2, 3, 10)

    def compute_errs(acc, plot=False):
        # import matplotlib.pyplot as plt

        errs_a = []
        errs_b = []
        errs_c = []

        for n in ns:
            x = np.linspace(0.1, 1, int(n))
            dx = x[1] - x[0]

            D = Diff(0, dx, acc=acc)
            D2 = D**2
            DD = D * D

            f = x**8

            exact = 8 * 7 * x**6
            actual_a = D2(f)
            actual_b = D(D(f))
            actual_c = DD(f)

            err_a = np.max(abs((actual_a - exact) / exact))
            err_b = np.max(abs((actual_b - exact) / exact))
            err_c = np.max(abs((actual_c - exact) / exact))
            errs_a.append(err_a)
            errs_b.append(err_b)
            errs_c.append(err_c)

        # if plot:
        #     plt.loglog(ns, errs_a, "-x", label="D^2f")
        #     plt.loglog(ns, errs_b, "-", label="D(D(f))")
        #     plt.loglog(ns, errs_c, "-", label="D*D*f")
        #     plt.grid()
        #     plt.show()

        slope_a = abs(loglog_slope(ns, errs_a))
        slope_b = abs(loglog_slope(ns, errs_b))
        slope_c = abs(loglog_slope(ns, errs_c))
        return slope_a, slope_b, slope_c

    slope_a, slope_b, slope_c = compute_errs(2, plot=False)

    assert abs(slope_a - 2) < 0.2
    # applying operators iteratively should reduce the order by one at a time:
    assert abs(slope_b - 1) < 0.2
    assert abs(slope_c - 2) < 0.2

    slope_a, slope_b, slope_c = compute_errs(4, plot=False)
    assert abs(slope_a - 4) < 0.2
    # applying operators iteratively should reduce the order by one at a time:
    assert abs(slope_b - 3) < 0.4
    assert abs(slope_c - 4) < 0.2


def loglog_slope(x, y):
    slope, intercept = np.polyfit(np.log(x), np.log(y), 1)
    return slope



"""
findiff is a Python package for finite difference numerical derivatives
and partial differential equations in any number of dimensions.

Features:

- Differentiate arrays of any number of dimensions along any axis with any desired accuracy order
- Accurate treatment of grid boundary
- Includes standard operators from vector calculus like gradient, divergence and curl
- Can handle uniform and non-uniform grids
- Can handle arbitrary linear combinations of derivatives with constant and variable coefficients
- Fully vectorized for speed
- Calculate raw finite difference coefficients for any order and accuracy for uniform and non-uniform grids
- Generate matrix representations of arbitrary linear differential operators
- Solve partial differential equations with Dirichlet or Neumann boundary conditions
- Generate differential operators for generic stencils
- Create symbolic representations of finite difference schemes
- Version 0.11.*: Completely remodeled API (backward compatibility is maintained, though)
- Version 0.12.*: Periodic boundary conditions for differential operators and PDEs
"""

__version__ = "0.12.2"


from .coefs import coefficients
from .compatible import Coef, Coefficient, FinDiff, Id
from .interface import Diff
from .operators import Identity
from .pde import PDE, BoundaryConditions
from .symbolic import SymbolicDiff, SymbolicMesh
from .vector import Curl, Divergence, Gradient, Laplacian

__all__ = [
    "coefficients",
    "Coef",
    "Coefficient",
    "FinDiff",
    "Id",
    "Diff",
    "Identity",
    "PDE",
    "BoundaryConditions",
    "SymbolicDiff",
    "SymbolicMesh",
    "Curl",
    "Divergence",
    "Gradient",
    "Laplacian",
]



import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from findiff import Gradient, Divergence, Curl, Laplacian


class TestGradient:

    def test_3d_gradient_on_scalar_func(self):
        axes, h, [X, Y, Z] = init_mesh(3, (50, 50, 50))
        f = np.sin(X) * np.sin(Y) * np.sin(Z)
        grad_f_ex = np.array(
            [
                np.cos(X) * np.sin(Y) * np.sin(Z),
                np.sin(X) * np.cos(Y) * np.sin(Z),
                np.sin(X) * np.sin(Y) * np.cos(Z),
            ]
        )
        grad = Gradient(spac=h, acc=4)
        grad_f = grad(f)
        assert_array_almost_equal(grad_f, grad_f_ex)

    def test_spacing_with_h(self):
        axes, h, [X, Y, Z] = init_mesh(3, (50, 50, 50))
        f = np.sin(X) * np.sin(Y) * np.sin(Z)
        grad_f_ex = np.array(
            [
                np.cos(X) * np.sin(Y) * np.sin(Z),
                np.sin(X) * np.cos(Y) * np.sin(Z),
                np.sin(X) * np.sin(Y) * np.cos(Z),
            ]
        )
        grad = Gradient(h=h, acc=4)
        grad_f = grad(f)
        assert_array_almost_equal(grad_f, grad_f_ex)

    def test_3d_gradient_on_scalar_func_non_uni(self):
        axes, h, [X, Y, Z] = init_mesh(3, (50, 50, 50))
        f = np.sin(X) * np.sin(Y) * np.sin(Z)
        grad_f_ex = np.array(
            [
                np.cos(X) * np.sin(Y) * np.sin(Z),
                np.sin(X) * np.cos(Y) * np.sin(Z),
                np.sin(X) * np.sin(Y) * np.cos(Z),
            ]
        )
        grad = Gradient(coords=axes, acc=4)
        grad_f = grad(f)
        assert_array_almost_equal(grad_f, grad_f_ex)

    def test_3d_gradient_on_vector_func_should_fail(self):
        axes, h, [X, Y, Z] = init_mesh(3, (50, 50, 50))
        f = np.array(
            [np.sin(X) * np.sin(Y) * np.sin(Z), np.sin(X) * np.sin(Y) * np.sin(Z)]
        )
        grad = Gradient(spac=h, acc=4)
        with pytest.raises(ValueError):
            grad(f)


class TestDivergence:

    def test_3d_divergence_on_vector_func(self):
        axes, h, [X, Y, Z] = init_mesh(3, (50, 50, 50))
        f = np.array([np.sin(X) * np.sin(Y) * np.sin(Z)] * 3)
        assert f.shape == (3, 50, 50, 50)
        div_f_ex = (
            np.cos(X) * np.sin(Y) * np.sin(Z)
            + np.sin(X) * np.cos(Y) * np.sin(Z)
            + np.sin(X) * np.sin(Y) * np.cos(Z)
        )

        div = Divergence(spac=h, acc=4)
        div_f = div(f)
        assert_array_almost_equal(div_f, div_f_ex)

    def test_3d_divergence_on_vector_func_of_wrong_dim(self):
        axes, h, [X, Y, Z] = init_mesh(3, (50, 50, 50))
        f = np.array([np.sin(X) * np.sin(Y) * np.sin(Z)] * 3)
        assert f.shape == (3, 50, 50, 50)
        div = Divergence(spac=[1, 1], acc=4)
        with pytest.raises(ValueError):
            div(f)


class TestCurl:

    def test_curl_on_3d_vector_func(self):
        axes, h, [X, Y, Z] = init_mesh(3, (50, 50, 50))
        f = np.array([np.sin(X) * np.sin(Y) * np.sin(Z)] * 3)
        assert f.shape == (3, 50, 50, 50)
        curl_f_ex = np.array(
            [
                np.sin(X) * np.cos(Y) * np.sin(Z) - np.sin(X) * np.sin(Y) * np.cos(Z),
                np.sin(X) * np.sin(Y) * np.cos(Z) - np.cos(X) * np.sin(Y) * np.sin(Z),
                np.cos(X) * np.sin(Y) * np.sin(Z) - np.sin(X) * np.cos(Y) * np.sin(Z),
            ]
        )
        curl = Curl(spac=h, acc=4)
        curl_f = curl(f)
        assert_array_almost_equal(curl_f, curl_f_ex)

    def test_curl_for_2d(self):
        axes, h, [X, Y, Z] = init_mesh(3, (50, 50, 50))
        f = np.array([np.sin(X) * np.sin(Y) * np.sin(Z)] * 3)
        assert f.shape == (3, 50, 50, 50)
        with pytest.raises(ValueError):
            Curl(spac=[1, 1], acc=4)

    def test_curl_for_2d_function(self):
        axes, h, [X, Y] = init_mesh(2, (50, 50))
        f = np.array([np.sin(X) * np.sin(Y)] * 2)
        curl = Curl(spac=[1, 1, 1], acc=4)
        with pytest.raises(ValueError):
            curl(f)

    def test_laplacian(self):
        axes, h, [X, Y] = init_mesh(2, (50, 50))
        f = X**2 + Y**2
        laplace = Laplacian(h)
        np.testing.assert_array_almost_equal(laplace(f), 4 * np.ones_like(X))


def init_mesh(ndims, npoints):
    axes = [np.linspace(-1, 1, npoints[k]) for k in range(ndims)]
    h = [x[1] - x[0] for x in axes]
    mesh = np.meshgrid(*axes, indexing="ij")
    return axes, h, mesh



from math import log

import numpy as np
import pytest

from findiff import FinDiff


class TestScaling:

    def fit_1d(self, acc):
        nx_list = 10 ** np.linspace(1.75, 2.0, 10)
        Lx = np.pi

        log_err_list = []
        log_dx_list = []

        for nx in nx_list:
            nx = int(nx)
            x = np.linspace(0.0, Lx, nx)
            dx = x[1] - x[0]
            f = np.sin(x)
            d_dx = FinDiff(0, dx)
            fx = d_dx(f, acc=acc)
            fxe = np.cos(x)
            err = np.max(np.abs(fxe - fx))
            log_dx_list.append(log(dx))
            log_err_list.append(log(err))

        fit = np.polyfit(log_dx_list, log_err_list, deg=1)
        return fit[0]

    def fit_2d(self, acc):
        nx_list = [10, 30, 100, 300]
        ny_list = [10, 30, 100, 300]
        Lx, Ly = 3, 3

        log_err_list = []
        log_dx_list = []

        for nx, ny in zip(nx_list, ny_list):
            x = np.linspace(0, Lx, nx)
            y = np.linspace(0, Ly, ny)
            dx, dy = x[1] - x[0], y[1] - y[0]
            X, Y = np.meshgrid(x, y, indexing="ij")
            f = np.sin(X) * np.sin(Y)
            d_dx = FinDiff(0, dx)
            fx = d_dx(f, acc=acc)
            fxe = np.cos(X) * np.sin(Y)
            err = np.max(np.abs(fxe - fx))
            log_dx_list.append(log(dx))
            log_err_list.append(log(err))

        fit = np.polyfit(log_dx_list, log_err_list, deg=1)
        return fit[0]

    @pytest.mark.parametrize("acc", [2, 4, 6])
    def test_1d_acc(self, acc):
        assert self.fit_1d(acc=acc) == pytest.approx(acc, 0.1)

    @pytest.mark.parametrize("acc", [2, 4])
    def test_2d_acc2(self, acc):
        assert self.fit_2d(acc=acc) == pytest.approx(acc, 0.1)

    def assertAlmostEqual(self, first, second, acc=None):
        if acc:
            assert first == pytest.approx(second, abs=acc)
        assert first == pytest.approx(second)



"""This module provides an interface to obsolete classes for backward compatibility."""

from findiff.interface import Diff
from findiff.operators import FieldOperator, Identity


def FinDiff(*args, **kwargs):
    r"""A representation of a general linear differential operator expressed in finite differences.

        FinDiff objects can be added with other FinDiff objects. They can be multiplied by
        objects of type Coefficient.

        FinDiff is callable, i.e. to apply the derivative, just call the object on the array to
        differentiate.

        :param args: variable number of tuples. Defines what derivative to take.
            If only one tuple is given, you can leave away the tuple parentheses.

        Each tuple has the form

               `(axis, spacing, count)`     for uniform grids

               `(axis, count)`              for non-uniform grids.

             `axis` is the dimension along which to take derivative.

             `spacing` is the grid spacing of the uniform grid along that axis.

             `count` is the order of the derivative, which is optional an defaults to 1.


        :param kwargs:  variable number of keyword arguments

            Allowed keywords:

            `acc`:    even integer
                  The desired accuracy order. Default is acc=2.

        This class is actually deprecated and will be replaced by the Diff class in the future.

    **Example**:


       For this example, we want to operate on some 3D array f:

       >>> import numpy as np
       >>> x, y, z = [np.linspace(-1, 1, 100) for _ in range(3)]
       >>> dx = x[1] - x[0]
       >>> dy = y[1] - y[0]
       >>> dz = z[1] - z[0]
       >>> X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
       >>> f = X**2 + Y**2 + Z**2

       To create :math:`\\frac{\\partial f}{\\partial x}` on a uniform grid with spacing dx, dy
       along the 0th axis or 1st axis, respectively, instantiate a FinDiff object and call it:

       >>> d_dx = FinDiff(0, dx)
       >>> d_dy = FinDiff(1, dy)
       >>> result = d_dx(f)

       For :math:`\\frac{\\partial^2 f}{\\partial x^2}` or :math:`\\frac{\\partial^2 f}{\\partial y^2}`:

       >>> d2_dx2 = FinDiff(0, dx, 2)
       >>> d2_dy2 = FinDiff(1, dy, 2)
       >>> result_2 = d2_dx2(f)
       >>> result_3 = d2_dy2(f)

       For :math:`\\frac{\\partial^4 f}{\partial x \\partial^2 y \\partial z}`, do:

       >>> op = FinDiff((0, dx), (1, dy, 2), (2, dz))
       >>> result_4 = op(f)


    """
    if len(args) > 3:
        raise ValueError("FinDiff accepts not more than 3 positional arguments.")

    def diff_from_tuple(tpl):
        if len(tpl) == 3:
            axis, h, order = tpl
            return Diff(axis, h, **kwargs) ** order
        elif len(tpl) == 2:
            axis, h = tpl
            return Diff(axis, h, **kwargs)

    if isinstance(args[0], (list, tuple)):
        diffs = []
        for tpl in args:
            diffs.append(diff_from_tuple(tpl))
        fd = diffs[0]
        for diff in diffs[1:]:
            fd = fd * diff
        return fd

    return diff_from_tuple(args)


###
### Define aliasses for backward compatibility:
###


class Coefficient(FieldOperator):
    pass


Coef = Coefficient
Id = Identity



import itertools
from itertools import product

import numpy as np


def interior_mask_as_ndarray(shape):
    ndims = len(shape)
    mask = np.zeros(shape, dtype=bool)
    mask[tuple([slice(1, -1)] * ndims)] = True
    return mask


def all_index_tuples_as_list(shape):
    ndims = len(shape)
    return list(product(*tuple([list(range(shape[k])) for k in range(ndims)])))


def get_long_indices_for_all_grid_points_as_ndarray(shape):
    return get_long_indices_for_all_grid_points_as_1d_array(shape).reshape(shape)


def get_long_indices_for_all_grid_points_as_1d_array(shape):
    return np.arange(np.prod(shape), dtype=np.int64)


def to_long_index(idx, shape):

    ndims = len(shape)
    long_idx = 0
    siz = 1
    for axis in range(ndims):
        idx_ = idx[ndims - 1 - axis]
        long_idx += idx_ * siz
        siz *= shape[ndims - 1 - axis]

    return long_idx


def to_index_tuple(long_idx, shape):
    ndims = len(shape)
    idx = np.zeros(ndims)
    for k in range(ndims):
        s = np.prod(shape[k + 1 :])
        idx[k] = long_idx // s
        long_idx = long_idx - s * idx[k]

    return tuple(idx)


def get_list_of_multiindex_tuples(shape):
    short_inds = [np.arange(shape[k]) for k in range(len(shape))]
    short_inds = list(itertools.product(*short_inds))
    return short_inds


#
# The following is working, but unused yet. Commented because there are no tests yet.
#
# def deprecated(reason="This feature is deprecated."):
#     def decorator(func_or_class):
#         if isinstance(func_or_class, type):  # Handle classes
#             original_init = func_or_class.__init__
#
#             @wraps(original_init)
#             def new_init(self, *args, **kwargs):
#                 warnings.warn(
#                     f"{func_or_class.__name__} is deprecated and will be removed in future versions: {reason}",
#                     category=DeprecationWarning,
#                     stacklevel=2,
#                 )
#                 original_init(self, *args, **kwargs)
#
#             func_or_class.__init__ = new_init
#             return func_or_class
#
#         # Handle functions
#         @wraps(func_or_class)
#         def wrapped(*args, **kwargs):
#             warnings.warn(
#                 f"{func_or_class.__name__} is deprecated: {reason}",
#                 category=DeprecationWarning,
#                 stacklevel=2,
#             )
#             return func_or_class(*args, **kwargs)
#
#         return wrapped
#
#     return decorator



# https://nbclient.readthedocs.io/en/latest/client.html

from pathlib import Path

import nbclient
import nbformat
import pytest


@pytest.mark.parametrize(
    "filename",
    [str(fp) for fp in Path("examples/").glob("**/*.ipynb")],
)
@pytest.mark.filterwarnings("ignore:Proactor event loop:RuntimeWarning")
def test_example_notebooks(filename):
    nb = nbformat.read(filename, as_version=4)
    client = nbclient.NotebookClient(nb, timeout=600, kernel_name="python3")
    client.execute()



import pytest

from findiff.grids import EquidistantAxis, make_grid, make_axis


def test_create_equidistantaxis():
    axis = EquidistantAxis(1, 0.1)
    assert axis.spacing == 0.1
    assert axis.dim == 1

    assert axis.periodic is False

    axis = EquidistantAxis(1, 0.1, periodic=True)
    assert axis.periodic is True


def test_create_equidistantaxis_fails_invalid_args():
    with pytest.raises(ValueError):
        EquidistantAxis(1, 0.0)

    with pytest.raises(ValueError):
        EquidistantAxis(-1, 0.1)

    with pytest.raises(ValueError, match="Dimension must be an integer"):
        EquidistantAxis(0.1, 1)


def test_makegrid_with_none_returns_none():
    assert make_grid(None) is None


def test_makegrid_with_invalid_arg_raises_exception():
    with pytest.raises(TypeError):
        make_grid("sndljas")


def test_makeaxis_with_gridaxis_returns_axis():
    axis = EquidistantAxis(1, 1)
    assert make_axis(1, axis) is axis



import numbers

import numpy as np


class GridAxis:

    def __init__(self, dim: int, periodic=False):
        if not isinstance(dim, numbers.Number) or int(dim) != dim:
            raise ValueError("Dimension must be an integer")
        if dim < 0:
            raise ValueError("Dimension must be >= 0.")
        self.dim = dim
        self.periodic = periodic


class EquidistantAxis(GridAxis):

    def __init__(self, dim: int, spacing: float, periodic=False):
        super().__init__(dim, periodic)
        if spacing <= 0:
            raise ValueError("Spacing must be > 0.")
        self.spacing = spacing


class NonEquidistantAxis(GridAxis):
    def __init__(self, dim: int, coords: np.ndarray, periodic=False):
        super().__init__(dim, periodic)
        self.coords = coords


class Grid:
    def __init__(self, *axes: GridAxis):
        self.axes = {ax.dim: ax for ax in axes}

    def get_axis(self, dim: int) -> GridAxis:
        return self.axes.get(dim)


def make_grid(config_or_grid):
    """Makes or returns a grid based on configuration or an actual Grid instance.

    Historically, the API allowed to specify grid using a variety of
    shortcuts using a single number, dicts or full Grid instances.

    The purpose of this function is to keep other modules closed with
    respect of addition an modification of GridAxis and Grid types.
    """
    if isinstance(config_or_grid, Grid):
        return config_or_grid
    elif isinstance(config_or_grid, dict):
        config = config_or_grid
        axes = []
        for dim, ax_config in config.items():
            if isinstance(ax_config, dict):
                ax = EquidistantAxis(
                    dim, ax_config["h"], periodic=ax_config.get("periodic", False)
                )
            else:
                ax = EquidistantAxis(dim, ax_config)
            axes.append(ax)
        return Grid(*axes)
    elif config_or_grid is None:
        return None
    else:
        raise TypeError(f"Unsupported grid type: {type(config_or_grid)}")


def make_axis(dim, config_or_axis, periodic=False):
    """Makes or returns a grid axis based on configuration or an actual GridAxis instance.

    Historically, the API allowed to specify axes using a variety of
    shortcuts using a single number, dicts or full GridAxis instances.

    The purpose of this function is to keep other modules closed with
    respect of addition an modification of GridAxis and Grid types.
    """

    if isinstance(config_or_axis, GridAxis):
        return config_or_axis
    if isinstance(config_or_axis, numbers.Number):
        return EquidistantAxis(dim, spacing=config_or_axis, periodic=periodic)
    elif isinstance(config_or_axis, np.ndarray):
        return NonEquidistantAxis(dim, coords=config_or_axis, periodic=periodic)



import numpy as np
import pytest
from sympy import Rational

from findiff import coefficients
from findiff.coefs import calc_coefs
from findiff.coefs import coefficients_non_uni


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_order2_acc2(analytic_inv):

    for analytic_inv in [True, False]:
        c = coefficients(deriv=2, acc=2, analytic_inv=analytic_inv)

        coefs = c["center"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([1.0, -2.0, 1.0]), coefs)
        offs = c["center"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([-1, 0, 1]), offs)

        coefs = c["forward"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array([2, -5, 4, -1]), coefs)
        offs = c["forward"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([0, 1, 2, 3]), offs)

        coefs = c["backward"]["coefficients"]
        np.testing.assert_array_almost_equal(np.array(([2, -5, 4, -1])[::-1]), coefs)
        offs = c["backward"]["offsets"]
        np.testing.assert_array_almost_equal(np.array([-3, -2, -1, 0]), offs)


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_order1_acc2(analytic_inv):

    c = coefficients(deriv=1, acc=2, analytic_inv=analytic_inv)

    coefs = c["center"]["coefficients"]
    np.testing.assert_array_almost_equal(np.array([-0.5, 0, 0.5]), coefs)
    offs = c["center"]["offsets"]
    np.testing.assert_array_almost_equal(np.array([-1, 0, 1]), offs)

    coefs = c["forward"]["coefficients"]
    np.testing.assert_array_almost_equal(np.array([-1.5, 2, -0.5]), coefs)
    offs = c["forward"]["offsets"]
    np.testing.assert_array_almost_equal(np.array([0, 1, 2]), offs)

    coefs = c["backward"]["coefficients"]
    np.testing.assert_array_almost_equal(-np.array([-1.5, 2, -0.5])[::-1], coefs)
    offs = c["backward"]["offsets"]
    np.testing.assert_array_almost_equal(np.array([-2, -1, 0]), offs)


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_order1_acc4(analytic_inv):
    c = coefficients(deriv=1, acc=4, analytic_inv=analytic_inv)
    coefs = c["center"]["coefficients"]
    np.testing.assert_array_almost_equal(
        np.array([1 / 12, -2 / 3, 0, 2 / 3, -1 / 12]), coefs
    )
    offs = c["center"]["offsets"]
    np.testing.assert_array_almost_equal(np.array([-2, -1, 0, 1, 2]), offs)

    coefs = c["forward"]["coefficients"]
    np.testing.assert_array_almost_equal(
        np.array([-25 / 12, 4, -3, 4 / 3, -1 / 4]), coefs
    )
    offs = c["forward"]["offsets"]
    np.testing.assert_array_almost_equal(np.array([0, 1, 2, 3, 4]), offs)

    coefs = c["backward"]["coefficients"]
    np.testing.assert_array_almost_equal(
        -np.array([-25 / 12, 4, -3, 4 / 3, -1 / 4])[::-1], coefs
    )
    offs = c["backward"]["offsets"]
    np.testing.assert_array_almost_equal(-np.array([0, 1, 2, 3, 4])[::-1], offs)


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_order2_acc4(analytic_inv):

    c = coefficients(deriv=2, acc=4, analytic_inv=analytic_inv)
    coefs = c["center"]["coefficients"]
    np.testing.assert_array_almost_equal(
        np.array([-1 / 12, 4 / 3, -2.5, 4 / 3, -1 / 12]), coefs
    )
    offs = c["center"]["offsets"]
    np.testing.assert_array_almost_equal(np.array([-2, -1, 0, 1, 2]), offs)

    coefs = c["forward"]["coefficients"]
    np.testing.assert_array_almost_equal(
        np.array([15 / 4, -77 / 6, 107 / 6, -13, 61 / 12, -5 / 6]), coefs
    )
    offs = c["forward"]["offsets"]
    np.testing.assert_array_almost_equal(np.array([0, 1, 2, 3, 4, 5]), offs)

    coefs = c["backward"]["coefficients"]
    np.testing.assert_array_almost_equal(
        np.array([15 / 4, -77 / 6, 107 / 6, -13, 61 / 12, -5 / 6])[::-1],
        coefs,
    )
    offs = c["backward"]["offsets"]
    np.testing.assert_array_almost_equal(-np.array([0, 1, 2, 3, 4, 5])[::-1], offs)


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_accuracy_central_deriv2_acc2(analytic_inv):
    coefs = calc_coefs(2, [-1, 0, 1], analytic_inv=analytic_inv)
    assert 2 == coefs["accuracy"]


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_accuracy_central_deriv1_acc2(analytic_inv):
    coefs = calc_coefs(1, [-1, 0, 1], analytic_inv=analytic_inv)
    assert 2 == coefs["accuracy"]


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_accuracy_left1_right0_deriv1_acc1(analytic_inv):
    coefs = calc_coefs(1, [-1, 0], analytic_inv=analytic_inv)
    assert 1 == coefs["accuracy"]


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_accuracy_left0_right3_deriv1_acc3(analytic_inv):
    coefs = calc_coefs(2, [0, 1, 2, 3], analytic_inv=analytic_inv)
    assert 2 == coefs["accuracy"]


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_accuracy_from_offsets_symbolic1(analytic_inv):
    coefs = calc_coefs(2, [0, 1, 2, 3], symbolic=True, analytic_inv=analytic_inv)
    assert 2 == coefs["accuracy"]


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_accuracy_from_offsets_symbolic2(analytic_inv):
    coefs = calc_coefs(2, [-4, -2, 0, 2, 4], symbolic=True, analytic_inv=analytic_inv)
    assert 4 == coefs["accuracy"]


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_coefs_from_offsets(analytic_inv):
    coefs = calc_coefs(1, [-2, 0, 1], analytic_inv=analytic_inv)
    np.testing.assert_array_almost_equal(
        coefs["coefficients"], [-1.0 / 6, -0.5, 2.0 / 3]
    )


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_coefs_from_offsets_no_central_point(analytic_inv):
    coefs = calc_coefs(1, [-2, 1, 2], analytic_inv=analytic_inv)
    np.testing.assert_array_almost_equal(coefs["coefficients"], [-1.0 / 4, 0, 1.0 / 4])


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_coefs_from_offsets_not_enough_points(analytic_inv):
    with pytest.raises(ValueError):
        coefficients(2, offsets=[-2, 2], analytic_inv=False)


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_calc_coefs_symbolic(analytic_inv):
    coefs = calc_coefs(1, [-2, 0, 1], symbolic=True, analytic_inv=analytic_inv)
    expected = [Rational(-1, 6), Rational(-1, 2), Rational(2, 3)]
    np.testing.assert_array_almost_equal(coefs["coefficients"], expected)


@pytest.mark.parametrize("analytic_inv", [True, False])
def test_non_uniform(analytic_inv):

    x = np.linspace(0, 10, 100)
    dx = x[1] - x[0]

    c_uni = coefficients(deriv=2, acc=2, analytic_inv=analytic_inv)
    coefs_uni = c_uni["center"]["coefficients"] / dx**2

    c_non_uni = coefficients_non_uni(deriv=2, acc=2, coords=x, idx=5)
    coefs_non_uni = c_non_uni["coefficients"]

    np.testing.assert_array_almost_equal(coefs_non_uni, coefs_uni)


def test_invalid_acc_raises_exception():
    with pytest.raises(ValueError):
        coefficients(deriv=1, acc=3)
    with pytest.raises(ValueError):
        coefficients(deriv=1, acc=0)
    with pytest.raises(ValueError):
        coefficients_non_uni(1, 3, None, None)
    with pytest.raises(ValueError):
        coefficients_non_uni(1, 0, None, None)


def test_invalid_deriv_raises_exception():
    with pytest.raises(ValueError):
        coefficients(-1, 2)
    with pytest.raises(ValueError):
        coefficients_non_uni(-1, 2, None, None)



import numpy as np
from scipy import sparse

from findiff.coefs import coefficients, coefficients_non_uni
from findiff.grids import EquidistantAxis, GridAxis, NonEquidistantAxis
from findiff.utils import (
    get_list_of_multiindex_tuples,
    get_long_indices_for_all_grid_points_as_1d_array,
    get_long_indices_for_all_grid_points_as_ndarray,
    to_long_index,
)


def build_differentiator(order: int, axis: GridAxis, acc):
    if isinstance(axis, EquidistantAxis):
        if not axis.periodic:
            return _FinDiffUniform(axis.dim, order, axis.spacing, acc)
        else:
            return _FinDiffUniformPeriodic(axis.dim, order, axis.spacing, acc)
    elif isinstance(axis, NonEquidistantAxis):
        if not axis.periodic:
            return _FinDiffNonUniform(axis.dim, order, axis.coords, acc)
        else:
            raise NotImplementedError("Periodic nonuniform axes not yet implemented")
    else:
        raise TypeError("Unknown axis type.")


class _FinDiffBase:

    def __init__(self, axis, order):
        self.axis = axis
        self.order = order

    def guard_valid_target(self, f):
        try:
            f.shape[self.axis]
        except AttributeError as err:
            raise ValueError(
                "Diff objects can only be applied to arrays or evaluated(!) functions returning arrays"
            ) from err

        if np.issubdtype(f.dtype, np.integer):
            f = f.astype(np.float64)
        return f

    def apply_to_array(self, yd, y, weights, off_slices, ref_slice, dim):
        """Applies the finite differences only to slices along a given axis"""

        ndims = len(y.shape)

        all = slice(None, None, 1)

        ref_multi_slice = [all] * ndims
        ref_multi_slice[dim] = ref_slice

        for w, s in zip(weights, off_slices):
            off_multi_slice = [all] * ndims
            off_multi_slice[dim] = s
            if abs(1 - w) < 1.0e-14:
                yd[tuple(ref_multi_slice)] += y[tuple(off_multi_slice)]
            else:
                yd[tuple(ref_multi_slice)] += w * y[tuple(off_multi_slice)]

    def shift_slice(self, sl, off, max_index):

        if sl.start + off < 0 or sl.stop + off > max_index:
            raise IndexError("Shift slice out of bounds")

        return slice(sl.start + off, sl.stop + off, sl.step)

    def matrix(self, shape):
        siz = np.prod(shape)
        mat = sparse.lil_matrix((siz, siz))
        self.write_matrix_entries(mat, shape)
        return sparse.csr_matrix(mat)

    def write_matrix_entries(self, mat, shape):
        raise NotImplementedError


class _FinDiffUniform(_FinDiffBase):

    def __init__(self, axis, order, spacing, acc):
        super().__init__(axis, order)
        self.spacing = spacing
        self.acc = acc
        coef_schemes = coefficients(self.order, acc)
        self.forward = coef_schemes["forward"]
        self.backward = coef_schemes["backward"]
        self.center = coef_schemes["center"]

    def __call__(self, f):
        f = self.guard_valid_target(f)

        npts = f.shape[self.axis]
        fd = np.zeros_like(f)
        num_bndry_points = len(self.center["coefficients"]) // 2

        self._apply_central_coefs(f, fd, npts, num_bndry_points)

        self._apply_forward_coefs(f, fd, npts, num_bndry_points)

        self._apply_backward_coefs(f, fd, npts, num_bndry_points)

        h_inv = 1.0 / self.spacing**self.order
        return fd * h_inv

    def _apply_backward_coefs(self, f, fd, npts, num_bndry_points):
        weights = self.backward["coefficients"]
        offsets = self.backward["offsets"]
        ref_slice = slice(npts - num_bndry_points, npts, 1)
        off_slices = [
            self.shift_slice(ref_slice, offsets[k], npts) for k in range(len(offsets))
        ]
        self.apply_to_array(fd, f, weights, off_slices, ref_slice, self.axis)

    def _apply_forward_coefs(self, f, fd, npts, num_bndry_points):
        weights = self.forward["coefficients"]
        offsets = self.forward["offsets"]
        ref_slice = slice(0, num_bndry_points, 1)
        off_slices = [
            self.shift_slice(ref_slice, offsets[k], npts) for k in range(len(offsets))
        ]
        self.apply_to_array(fd, f, weights, off_slices, ref_slice, self.axis)

    def _apply_central_coefs(self, f, fd, npts, num_bndry_points):
        weights = self.center["coefficients"]
        offsets = self.center["offsets"]
        ref_slice = slice(num_bndry_points, npts - num_bndry_points, 1)
        off_slices = [
            self.shift_slice(ref_slice, offsets[k], npts) for k in range(len(offsets))
        ]
        self.apply_to_array(fd, f, weights, off_slices, ref_slice, self.axis)

    def write_matrix_entries(self, mat, shape):
        long_indices_nd = get_long_indices_for_all_grid_points_as_ndarray(shape)
        for scheme in ["center", "forward", "backward"]:

            offsets_long = self._convert_1D_offsets_to_long_indices(
                self.axis, getattr(self, scheme)["offsets"], shape
            )

            multi_slice = self._get_multislice_for_scheme(self.axis, scheme, shape)
            Is = long_indices_nd[tuple(multi_slice)].reshape(-1)

            coefs = getattr(self, scheme)["coefficients"]
            for o, c in zip(offsets_long, coefs):
                v = c / self.spacing**self.order
                mat[Is, Is + o] = v

    def _get_multislice_for_scheme(self, axis, scheme, shape):
        ndims = len(shape)
        multi_slice = [slice(None, None)] * ndims
        nside = len(self.center["coefficients"]) // 2
        if scheme == "center":
            multi_slice[axis] = slice(nside, -nside)
        elif scheme == "forward":
            multi_slice[axis] = slice(0, nside)
        else:
            multi_slice[axis] = slice(-nside, None)
        return multi_slice

    def _convert_1D_offsets_to_long_indices(self, axis, offsets_1d, shape):
        ndims = len(shape)
        offsets_long = []
        for o_1d in offsets_1d:
            o_nd = np.zeros(ndims, dtype=int)
            o_nd[axis] = o_1d
            o_long = to_long_index(o_nd, shape)
            offsets_long.append(o_long)
        return offsets_long


class _FinDiffUniformPeriodic(_FinDiffBase):

    def __init__(self, axis, order, spacing, acc):
        super().__init__(axis, order)
        self.spacing = spacing
        self.acc = acc
        self.coefs = coefficients(self.order, acc)["center"]

    def __call__(self, f):
        f = self.guard_valid_target(f)

        fd = np.zeros_like(f)
        for off, coef in zip(self.coefs["offsets"], self.coefs["coefficients"]):
            fd += coef * np.roll(f, -off, axis=self.axis)
        h_inv = 1.0 / self.spacing**self.order
        return fd * h_inv

    def write_matrix_entries(self, mat, shape):
        Is = get_long_indices_for_all_grid_points_as_1d_array(shape)
        h_inv = 1 / self.spacing**self.order
        for o, c in zip(self.coefs["offsets"], self.coefs["coefficients"]):
            Is_off = self._get_offset_indices_long(o, shape)
            mat[Is, Is_off] = c * h_inv

    def _get_offset_indices_long(self, o, shape):
        ndims = len(shape)
        idxs_short = [np.arange(n) for n in shape]
        idxs_short[self.axis] = np.roll(np.arange(shape[self.axis]), -o)
        grid = np.meshgrid(*idxs_short, indexing="ij")
        index_tuples = np.stack(grid, axis=-1).reshape(-1, ndims)
        Is_off = np.ravel_multi_index(index_tuples.T, shape)
        return Is_off


class _FinDiffNonUniform(_FinDiffBase):
    def __init__(self, axis, order, coords, acc):
        super().__init__(axis, order)
        self.coords = coords
        self.acc = acc
        self.coef_list = [
            coefficients_non_uni(order, self.acc, self.coords, i)
            for i in range(len(coords))
        ]

    def __call__(self, y):
        """The core function to take a partial derivative on a non-uniform grid"""
        y = self.guard_valid_target(y)

        dim = self.axis

        yd = np.zeros_like(y)

        ndims = len(y.shape)
        multi_slice = [slice(None, None)] * ndims
        ref_multi_slice = [slice(None, None)] * ndims

        for i, _ in enumerate(self.coords):

            coefs = self.coef_list[i]
            ref_multi_slice[dim] = i

            for off, w in zip(coefs["offsets"], coefs["coefficients"]):
                multi_slice[dim] = i + off
                yd[tuple(ref_multi_slice)] += w * y[tuple(multi_slice)]

        return yd

    def write_matrix_entries(self, mat, shape):

        coords = self.coords

        short_inds = get_list_of_multiindex_tuples(shape)

        coef_dicts = []
        for i in range(len(coords)):
            coef_dicts.append(coefficients_non_uni(self.order, self.acc, coords, i))

        long_inds = get_long_indices_for_all_grid_points_as_ndarray(shape)
        for base_ind_long, base_ind_short in enumerate(short_inds):
            cd = coef_dicts[base_ind_short[self.axis]]
            cs, os = cd["coefficients"], cd["offsets"]
            for c, o in zip(cs, os):
                off_short = np.zeros(len(shape), dtype=int)
                off_short[self.axis] = int(o)
                off_ind_short = np.array(base_ind_short, dtype=int) + off_short
                off_long = long_inds[tuple(off_ind_short)]

                mat[base_ind_long, off_long] += c
